# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright NI SP GmbH. All Rights Reserved.

# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#
# Build NICE DCV in a Container including GPU 3D support 
#
# More information : https://www.ni-sp.com/nice-dcv-in-containers/
#

target="dcv-centos7-gpu"

RED='\033[0;31m'; GREEN='\033[0;32m'; GREY='\033[0;37m'; BLUE='\034[0;37m'; NC='\033[0m' 
ORANGE='\033[0;33m'; BLUE='\033[0;34m';
echo -e "${GREEN}###################################################"
echo -e "NICE DCV for Container Installation starting"
echo -e "  created by NI SP GmbH                         " 
echo -e "  https://www.ni-sp.com/contact/              " 
echo -e "###################################################${NC}"
echo
sleep 1.5
    
# Prepare the NICE DCV license in case needed
# Please contact NI SP GmbH in case you need a NICE DCV trial license at https://www.ni-sp.com/contact/. 
# On AWS you can enable license access according to this guide:
# https://docs.aws.amazon.com/dcv/latest/adminguide/setting-up-license.html

[ -f /etc/os-release ] && . /etc/os-release
if [ -f /etc/redhat-release -o "$ID" == "ubuntu" -o "$ID" == "amzn" ] ; then
    :
else
    if [ -f /usr/bin/nvidia-smi ] ; then
	:   # we seem to be good 
    else
	echo -e "${RED}##########################################"
	echo -e "Please install the nVidia driver first with the same version as in the Dockerfile"
	echo -e "At the moment we support only Redhat/Centos or Ubuntu nvidia driver installation on the host as part of this script"
	echo -e "##########################################${NC}"
	exit
    fi
fi    
ip=`curl --silent http://169.254.169.254/latest/meta-data/local-ipv4 | grep \.`
if [ "$ip" == "" ] ; then
    # We are not on AWS and need a DCV trial license
    onaws="0"
    license="license.lic"
    if [ -f $license ] ; then
	# great, we have a license
	:
    else
	echo 
	echo -n "Please specify the path to the NICE DCV trial license (e.g. license.lic) : "
	read license
    fi
    # check license
    res=`grep "FEATURE nice" $license | head -1`
    if [ ! -f $license -o  "$res" == "" ]; then
	echo 
	echo "It seems there is no valid NICE DCV trial license ... "
	echo "Please contact NI SP GmbH for a NICE DCV trial license at https://www.ni-sp.com/contact/." 
	echo
	exit
    fi
    [ ! -f license.lic ] && cp $license license.lic   # copy the license file to a standard license name 
else
    # on AWS NICE DCV please enable license access to the S3 bucket 
    touch license.lic
    onaws="1"
fi

# Install nVidia driver on host
smi_res=""
if [ -f /usr/bin/nvidia-smi ] ; then
    smi_res=`nvidia-smi --query-gpu=gpu_name --format=csv,noheader --id=0 | sed -e 's/ /-/g' 2>&1`
fi

if [ "$smi_res" == "" -o "`echo $smi_res | grep 'NVIDIA-SMI has failed'`" != "" ] ; then
    echo -e "${GREEN}##########################################"
    echo -e "Starting nVidia driver installation "
    echo -e "##########################################${NC}"
    sleep 0.75
    if [ "`grep "blacklist nouveau" /etc/modprobe.d/blacklist.conf 2>&1 | grep -v 'No such' `" == "" ] ; then
	echo 
	echo -n "We will first disable the nouveau driver and reboot the server. Is this ok ? (y/n) "
	read answer
	if [ "$answer" == "y" -o "$answer" == "yes" ] ; then
	    :
	else
	    echo "Exiting ..."
	    exit
	fi
        echo -e "${GREEN}##########################################"
	echo -e "Disabling the nouveau driver"
	echo -e "##########################################${NC}"
	cat << EOF | sudo tee --append /etc/modprobe.d/blacklist.conf
blacklist nouveau
options nouveau modeset=0
EOF
	echo 'GRUB_CMDLINE_LINUX="rdblacklist=nouveau"' | sudo tee -a /etc/default/grub > /dev/null
	if [ "$ID" == "ubuntu" ] ; then
	    sudo update-grub
	else
	    sudo yum update -y
	    sudo grub2-mkconfig -o /boot/grub2/grub.cfg
	fi
	echo 
	echo "Rebooting ..."
	echo -e "${GREEN}##########################################"
	echo "Please relogin to your server and restart "
	echo "   the build script $0 "
	echo "   to continue the installation." 
	echo -e "##########################################${NC}"
	sudo reboot # to remove nouveau driver in case
    else
        echo -e "${GREEN}"
	echo "Nouveau driver is disabled - continuing ..."
	echo -e    "${NC}"
    fi
    # Install nVidia driver
    echo -e "${GREEN}##########################################"
    echo "Installing nVidia driver on the host"
    echo -e "##########################################${NC}"
    sleep 1
    distribution=$(. /etc/os-release;echo $ID$VERSION_ID)
    if [ "$ID" == "ubuntu" ] ; then
	# install nVidia driver for "ubuntu"
	sudo apt-get update -y
	sudo apt-get install -y gcc make linux-headers-$(uname -r)
	if [ "$onaws" == "0" ] ; then 
	    wget -q http://us.download.nvidia.com/XFree86/Linux-x86_64/430.26/NVIDIA-Linux-x86_64-430.26.run -O /tmp/NVIDIA-installer.run
	else
	    sudo apt install -y awscli
	    # pip3 install awscli --upgrade --user # --force-reinstall 
	    aws s3 cp --no-sign-request --recursive s3://ec2-linux-nvidia-drivers/latest/ .
	    # e.g. ./NVIDIA-Linux-x86_64-470.82.01-grid-aws.run
	    mv ./NVIDIA-Linux-x86_64*.run /tmp/NVIDIA-installer.run
	fi
	
    else
	if [ "$ID" == "amzn" ] ; then
	    sudo yum install -y gcc kernel-devel-$(uname -r) 
	else
	    sudo yum -y install kernel-devel-`uname -r` kernel-headers-`uname -r`
	    if [ "$distribution" == "centos7" ] ; then
		sudo yum -y install https://dl.fedoraproject.org/pub/epel/epel-release-latest-7.noarch.rpm
	    else
		sudo yum -y install https://dl.fedoraproject.org/pub/epel/epel-release-latest-8.noarch.rpm
	    fi
	fi
	sudo yum -y install dkms wget vim
	if [ "$onaws" == "0" ] ; then 
	    wget -q http://us.download.nvidia.com/XFree86/Linux-x86_64/430.26/NVIDIA-Linux-x86_64-430.26.run -O /tmp/NVIDIA-installer.run
	else
	    sudo yum install python3-pip -y
	    pip3 install awscli --upgrade --user # --force-reinstall 
	    aws s3 cp --no-sign-request --recursive s3://ec2-linux-nvidia-drivers/latest/ .
	    # e.g. ./NVIDIA-Linux-x86_64-470.82.01-grid-aws.run
	    mv ./NVIDIA-Linux-x86_64*.run /tmp/NVIDIA-installer.run && chmod o+x /tmp/NVIDIA-installer.run
	fi
    fi
    if [ "$ID" == "amzn" -a "`uname -r | grep 5.10`" != "" ] ; then
	sudo CC=/usr/bin/gcc10-cc /tmp/NVIDIA-installer.run --accept-license \
             --no-questions   --no-backup    --ui=none
    else
	sudo bash /tmp/NVIDIA-installer.run --accept-license \
             --no-questions   --no-backup    --ui=none
    fi
    mv /tmp/NVIDIA-installer.run .    # we need it later for the container
    echo -e "${GREEN}##########################################"
    echo "Finished nVidia driver installation "
    echo -e "##########################################${NC}"
    sleep 0.75
fi

# On Centos we use podman, on Ubuntu docker

if [ -f /etc/redhat-release ] ; then
    sudo yum install podman container-selinux -y
    CTRMGR="podman"
    # https://docs.nvidia.com/datacenter/cloud-native/container-toolkit/install-guide.html#podman
    echo -e "${GREEN}##########################################"
    echo "Installing nVidia container environment "
    echo -e "##########################################${NC}"
    distribution=$(. /etc/os-release;echo $ID$VERSION_ID) \
	&& curl -s -L https://nvidia.github.io/nvidia-docker/$distribution/nvidia-docker.repo | sudo tee /etc/yum.repos.d/nvidia-docker.repo
    sudo yum clean expire-cache
    sudo yum install -y nvidia-container-toolkit
    echo -e "${GREEN}##########################################"
    echo "Finished nVidia container environment "
    echo -e "##########################################${NC}"
sleep 1
elif [ "$ID" == "ubuntu" ] ; then
    CTRMGR="docker"
    if [ "`which docker`" == "/usr/bin/docker" ] ; then
	:
    else
	# Install docker 
	curl --silent https://get.docker.com | sh  && sudo systemctl --now enable docker
    fi
    if [ "`apt list  nvidia-docker2 2>&1 | grep installed`" == "" ] ; then
	echo -e "${GREEN}##########################################"
	echo "Installing nVidia container environment "
	echo -e "##########################################${NC}"
	distribution=$(. /etc/os-release;echo $ID$VERSION_ID) \
	    && curl -s -L https://nvidia.github.io/nvidia-docker/gpgkey | sudo apt-key add - \
	    && curl -s -L https://nvidia.github.io/nvidia-docker/$distribution/nvidia-docker.list | sudo tee /etc/apt/sources.list.d/nvidia-docker.list
	sudo apt-get update
	sudo apt-get install -y nvidia-docker2
	sudo systemctl restart docker
	echo -e "${GREEN}##########################################"
	echo "Finished nVidia container environment "
	echo -e "##########################################${NC}"
sleep 1
    fi
    echo
elif [ "$ID" == "amzn" ] ; then
    CTRMGR="docker"
    if [ "`which docker`" != "/usr/bin/docker" ] ; then
	if [ "$VERSION" == "2" ] ; then
	    sudo amazon-linux-extras install docker -y
	    sudo systemctl enable docker
	else
	    sudo yum install docker -y
	fi
	sudo service docker start
	sudo usermod -a -G docker ec2-user
    fi
else
    echo "OS not detected, not clear how to work with containers ..."
    exit
fi


# In case you want to build the container as standard user please check https://github.com/containers/podman/blob/main/docs/tutorials/rootless_tutorial.md
# create namespaces so a standard user can build the container
# sudo sysctl user.max_user_namespaces=15000

# Build the container
echo -e "${GREEN}##########################################"
echo "NICE DCV Container Build starting ..."
echo -e "##########################################${NC}"
sleep 1

sudo $CTRMGR build -t "$target" .

echo -e "${GREEN}##########################################"
echo "NICE DCV Container Build finished." 
echo -e "##########################################${NC}"
sleep 2

# Run the container
# in case the nr of visible GPUs should be reduced you can use e.g.  "-e NVIDIA_VISIBLE_DEVICES=0,1,2"
if [ "$CTRMGR" == "docker" ] ; then
    # cmd="sudo $CTRMGR run  --gpus all --device=/dev/nvidia0 --device=/dev/nvidiactl --privileged --rm --network="host" $target &"
    if [ "$ID" == "amzn" ] ; then
	cmd="sudo $CTRMGR run  --privileged --rm --network="host" $target &"
    else
	cmd="sudo $CTRMGR run  --gpus all --privileged --rm --network="host" $target &"
    fi
else
    # podman
    # cmd="sudo $CTRMGR run --security-opt=label=disable --hooks-dir=/usr/share/containers/oci/hooks.d/ --device=/dev/nvidia0 --device=/dev/nvidiactl --privileged --rm --network="host" $target &" 
    cmd="sudo $CTRMGR run --security-opt=label=disable --hooks-dir=/usr/share/containers/oci/hooks.d/ --privileged --rm --network="host" $target &" 
fi
# check for running container
check="`sudo $CTRMGR ps | grep  $target`"
if [ "$check" != "" ] ; then
    id="`sudo $CTRMGR ps | grep -v CONT | awk '{print $1}' | head -1`" 
    echo ""
    echo -e "${ORANGE}You have a container \"$target\" running already. You can stop it with \"sudo $CTRMGR stop $id\".${NC}"
    echo -n "Do you want to create another DCV container anyway which might have some issues? (y/n) "
    read answer
    [ "$answer" != "y" ] && exit 
fi

eval $cmd
# echo Now you can connect to DCV ...
# echo
# echo The default user name is “user” and the password “dcv” which
# echo can be adapted in the “startup_script.sh”.
# echo
# echo "Web browser: https://External_IP_or_Server_Name:8443 \(accept the security exception as we do not have a SSL certificate installed\) – or –"
# echo
# echo
# echo "DCV native client: Enter “External_IP_or_Server_Name” in the connection field (the portable DCV client can be downloaded here: https://download.nice-dcv.com/)"
# echo

sleep 12

echo -e "${GREEN}##########################################"
echo "How to use the NICE DCV container"
echo -e "##########################################${NC}"
# echo To login find out the container id of the DCV container and run bash inside
container_id=`sudo $CTRMGR ps  | grep dcv | awk '{print $1}' | head -1`
if [ "$container_id" == "" ] ; then
    echo It seems the container did not start properly ... please check the messages above.
    echo Exiting ...
    exit
fi
echo -e "The NICE DCV container id is \"$container_id\".\nHere is the output of the command \"sudo $CTRMGR ps\": ${GREEN} "
sudo $CTRMGR ps 
echo -e "${NC}"
echo You can login to the container with the following command:
echo -e "     ${GREEN}sudo $CTRMGR exec -ti $container_id /bin/bash${NC}"
echo 
# sudo podman exec -ti $container_id /bin/bash
# inside the container you can run e.g.
# dcv list-sessions

# echo create new user user2
# _username="user2" ;   _passwd="dcv"
# adduser "${_username}" -G wheel; echo "${_username}:${_passwd}" |chpasswd
# /usr/bin/dcv create-session --storage-root=%home% --owner "${_username}" --user "${_username}" "${_username}session"
# dcv list-sessions

# Stop the DCV container
container_id=`sudo $CTRMGR ps  | grep dcv | awk '{print $1}' | head -1`
echo You can stop the container with the following command:
echo -e "     ${ORANGE}sudo $CTRMGR stop $container_id${NC}"
echo
echo You can run the NICE DCV container again with the command: 
echo -e "     ${GREEN}$cmd${NC}"
echo
echo -e "${GREEN}##########################################${NC}"
echo 
echo -e "${ORANGE}Please wait for the NICE DCV connection information to show up soon ...${NC}"
echo

# DEBUG
# echo "##########################################"
# echo "##########################################"
# echo sudo $CTRMGR exec -ti $container_id less /var/log/dcv/agent.user.usersession.log
# echo "##########################################"
# echo "##########################################"
